import IPython
from IPython.display import Audio, Video, YouTubeVideo, Code, Image
import sys


def audio(path: str):
    return IPython.display.display(Audio(path))


def video(path: str, embed=False):
    return IPython.display.display(Video(path, embed=embed))


def youtube_video(id_: str):
    return IPython.display.display(YouTubeVideo(id_))


def image(path: str):
    return IPython.display.display(Image(path))


def code(path: str, language: str = None):
    from pygments import highlight
    from pygments.formatters.html import HtmlFormatter
    # fix JupyterLab from https://github.com/ipython/ipython/issues/11747
    def _jupyterlab_repr_html_(self):
        fmt = HtmlFormatter()
        style = "<style>{}\n{}</style>".format(
            fmt.get_style_defs(".output_html"), fmt.get_style_defs(".jp-RenderedHTML")
        )
        return style + highlight(self.data, self._get_lexer(), fmt)

    IPython.display.Code._repr_html_ = _jupyterlab_repr_html_
    return IPython.display.display(Code(path, language=language))


class Display:
    audio = staticmethod(audio)
    video = staticmethod(video)
    image = staticmethod(image)
    youtube_video = staticmethod(youtube_video)
    code = staticmethod(code)

    def __call__(self, path: str, *args, **kwargs):
        ext = path.split(".")[-1]
        if ext in ("mp3", "wav", "ogg"):
            return self.audio(path)
        elif ext in ("mp4", "avi", "mkv"):
            return self.video(path, *args, **kwargs)
        elif ext in ("jpg", "jpeg", "png", "bmp", "gif"):
            return self.image(path)
        elif ext in ("py", "c", "cpp", "java", "js", "html", "css", "json", "yaml", "yml", "xml", "sh", "bat"):
            return self.code(path, *args, **kwargs)
        else:
            raise ValueError(f"Unknown file type: {ext}")


sys.modules[__name__] = Display()



import toml


class TomlWriter:

    @classmethod
    def to(cls, path: str, obj: dict):
        """
        Save as Toml file from a dictionary.

        Args:
            path (``str``): The path of output file.
            obj (``dict``): The toml data which need to be used for output.
        """
        with open(path, "w", encoding="UTF-8") as f:
            toml.dump(obj, f)

    @classmethod
    def __new__(cls, path: str, obj: dict):
        """Alias of :meth:`~tensorneko.io.toml.toml_writer.TomlWriter.to`."""
        cls.to(path, obj)



import glob
import os.path
from abc import ABC, abstractmethod

from numpy import ndarray

from ...backend._tqdm import import_tqdm_auto


class AbstractFaceDetector(ABC):

    @abstractmethod
    def detect_face(self, image: ndarray, *args, **kwargs):
        pass

    @abstractmethod
    def crop_image(self, image_path: str, out_path: str, *args, **kwargs) -> None:
        pass

    @abstractmethod
    def crop_video(self, video_path: str, out_path: str, *args, **kwargs) -> None:
        pass

    def crop_image_dir(self, image_dir: str, out_dir: str, pattern="*.jpg", progress_bar = False, *args, **kwargs
    ) -> None:
        all_images = glob.glob(os.path.join(image_dir, pattern), root_dir=image_dir)
        if progress_bar:
            all_images = import_tqdm_auto().tqdm(all_images)
        for image_path in all_images:
            out_path = os.path.join(out_dir, image_path)
            self.crop_image(image_path, out_path, *args, **kwargs)

    def crop_video_dir(self, video_dir: str, out_dir: str, pattern="*.mp4", progress_bar = False, *args, **kwargs
    ) -> None:
        all_videos = glob.glob(os.path.join(video_dir, pattern), root_dir=video_dir)
        if progress_bar:
            all_videos = import_tqdm_auto().tqdm(all_videos)
        for video_path in all_videos:
            out_path = os.path.join(out_dir, video_path)
            self.crop_video(video_path, out_path, *args, **kwargs)



from typing import Callable, Coroutine, Any

from .bus import EventBus
from .event import E


class SubscribeDecorator:

    def __call__(self, func: Callable[[E], None]):
        return EventBus.default.subscribe(func)

    def coro(self, func: Callable[[E], Coroutine[Any, Any, None]]):
        return EventBus.default.subscribe_async(func)

    def thread(self, func: Callable[[E], None]):
        return EventBus.default.subscribe_thread(func)

    def process(self, func: Callable[[E], None]):
        return EventBus.default.subscribe_process(func)


subscribe = SubscribeDecorator()



from contextlib import contextmanager
from typing import TypeVar

from .bus import EventBus


class EventMeta(type):

    def __call__(cls, *args, **kwargs):
        event = super().__call__(*args, **kwargs)
        event.bus = kwargs.get("bus", EventBus.default)
        event.bus.emit(event, blocking=_blocking_flag)
        return event


_blocking_flag = True


@contextmanager
def no_blocking():
    """
    Context manager to disable blocking event handler.
    """
    global _blocking_flag
    _blocking_flag = False
    yield
    _blocking_flag = True


class Event(metaclass=EventMeta):
    bus: EventBus

    def __init_subclass__(cls, **kwargs):
        if "bus" in cls.__init__.__annotations__:
            raise TypeError("`bus` parameter is preserved. It should not be annotated in the __init__ method")


E = TypeVar("E", bound=Event)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Union, Callable, List, Optional, Iterable

from ..monad.monad import Monad
from ...type import T, R
from ....backend._tqdm import import_tqdm_auto
from ....backend.parallel import ParallelType


class AbstractSeq(Iterable[T], Monad[T], ABC):

    @abstractmethod
    def __getitem__(self, item: Union[int, slice]) -> Union[T, AbstractSeq[T]]:
        ...

    @abstractmethod
    def __iter__(self):
        ...

    @abstractmethod
    def map(self, f: Callable[[T], R]) -> AbstractSeq[R]:
        ...

    @abstractmethod
    def for_each(self, f: Callable[[T], None], progress_bar: bool = False, parallel_type: Optional[ParallelType] = None
    ) -> None:
        ...

    @abstractmethod
    def with_for_each(self, f: Callable[[T], None], progress_bar: bool = False,
        parallel_type: Optional[ParallelType] = None
    ) -> AbstractSeq[T]:
        ...

    @abstractmethod
    def filter(self, f: Callable[[T], bool]) -> AbstractSeq[T]:
        ...

    @abstractmethod
    def reduce(self, f: Callable[[T, T], T]) -> T:
        ...

    @abstractmethod
    def flatten(self: AbstractSeq[AbstractSeq[T]]) -> AbstractSeq[T]:
        ...

    @abstractmethod
    def flat_map(self, f: Callable[[T], AbstractSeq[R]]) -> AbstractSeq[R]:
        ...

    @abstractmethod
    def skip(self, n: int) -> AbstractSeq[T]:
        ...

    @abstractmethod
    def take(self, n: int) -> AbstractSeq[T]:
        ...

    @property
    @abstractmethod
    def head(self) -> T:
        ...

    @property
    @abstractmethod
    def tail(self) -> AbstractSeq[T]:
        ...

    @abstractmethod
    def repeat(self, n: int) -> AbstractSeq[T]:
        ...

    @abstractmethod
    def to_list(self) -> List[T]:
        ...

    @property
    def _tqdm(self):
        return import_tqdm_auto().tqdm
